/***********************************************************************
 *
 * PROJECT: PalmXQ
 * FILE:    board.c
 * AUTHOR:  Cheok Yan Cheng
 *
 * DESCRIPTION: Perform all the board operation.
 *                   
 * LAST REVISION:   24/02/2002
 *
 ***********************************************************************/

#include <PalmOS.h>
#include "palmxqrsc.h"
#include "data.h"
#include "protos.h"
#include "defs.h"

#ifndef ARM
/***********************************************************************
 *
 * FUNCTION:    DrawBoard
 *
 * DESCRIPTION: Draws XiangQi board on gadget.
 *
 * PARAMETERS:  form        FormType pointer to the form which contains
 *                          gadget.
 *              gadgetID    ID of the gadget.
 *              piece       Piece type information array.
 *              color       Piece color information array.
 *              isFlip      Should the board be flipped.
 *
 *
 * RETURNED:    NONE
 *
 * COMMENT:     The size of piece abd color must >= BRD_SIZE. Otherwise,
 *              the behaviour of this function is undetermined.
 *
 ***********************************************************************/
void DrawBoard(FormType *form, UInt16 gadgetID,
            UInt8 *piece, UInt8 *color, Boolean isFlip)
{
    FormType *oldForm;
    UInt16 gadgetIndex;
    RectangleType boardRect;
    RectangleType gadgetRect;
    RectangleType offWinRect;
    WinHandle offWin;
    WinHandle oldWin;
    Int16 widthPerSqu;
    Int16 heightPerSqu;
    Int16 i;
    Int16 offX;
    Int16 offY;
    Int16 brdX;
    Int16 brdY;
    Err err;
    MemHandle bitmapH;
    BitmapType *bitmap;

    static Int16 bitmapID[2][7] = {
        BmpDarkPawn,
        BmpDarkAssistance,
        BmpDarkElephant,
        BmpDarkKnight,
        BmpDarkCannon,
        BmpDarkRook,
        BmpDarkKing,
        BmpLightPawn,
        BmpLightAssistance,
        BmpLightElephant,
        BmpLightKnight,
        BmpLightCannon,
        BmpLightRook,
        BmpLightKing };
            
    // Get the current drawing form
    oldForm = FrmGetActiveForm();
    FrmSetActiveForm(form);

    // Get the rectangle bound of gadget
    gadgetIndex = FrmGetObjectIndex(form, gadgetID);
    FrmGetObjectBounds(form, gadgetIndex, &gadgetRect);

    // Create offscreen window
    offWin = WinCreateOffscreenWindow(
            gadgetRect.extent.x,
            gadgetRect.extent.y,
            nativeFormat,
            &err);
    oldWin = WinSetDrawWindow(offWin);
    WinEraseWindow();

    widthPerSqu = (gadgetRect.extent.x - BRD_BOUND_PIXEL) / (BRD_WIDTH);
    heightPerSqu = (gadgetRect.extent.y - BRD_BOUND_PIXEL) / (BRD_HEIGHT);
    offX = widthPerSqu / 2;
    offY = heightPerSqu / 2;

    // Draw board on offscreen window
    for(i = 0; i < BRD_WIDTH; i++)
    {
        WinDrawLine(offX + widthPerSqu * i,
                    offY,
                    offX + widthPerSqu * i,

                    // Minus 1 since the screen coordinate is 0 based
                    gadgetRect.extent.y - offY - 1);    
    }

    for(i = 0; i < BRD_HEIGHT; i++)
    {
        WinDrawLine(offX,
                    offY + heightPerSqu * i,

                    // Minus 1 since the screen coordinate is 0 based
                    gadgetRect.extent.x - offX - 1,
                    offY + heightPerSqu * i);
    }

    for(i = 1; i < BRD_WIDTH - 1; i++)
    {
        WinEraseLine(offX + widthPerSqu * i,
                    offY + heightPerSqu * (BRD_HEIGHT / 2 - 1) + 1,
                    offX + widthPerSqu * i,
                    offY + heightPerSqu * BRD_HEIGHT / 2 - 1);
    }

    // Draw the boudary rectangle
    RctSetRectangle(&boardRect,
                    offX - 1,
                    offY - 1,
                    (BRD_WIDTH - 1) * widthPerSqu + 3,
                    (BRD_HEIGHT - 1) * heightPerSqu + 3);
    WinDrawRectangleFrame(simpleFrame, &boardRect);

    // Draw king palace
    WinDrawLine(offX + widthPerSqu * 3,
                offY,
                offX + widthPerSqu * 5,
                offY + heightPerSqu * 2);

    WinDrawLine(offX + widthPerSqu * 5,
                offY,
                offX + widthPerSqu * 3,
                offY + heightPerSqu * 2);

    WinDrawLine(offX + widthPerSqu * 3,
                offY + heightPerSqu * 7,
                offX + widthPerSqu * 5,
                offY + heightPerSqu * 9);

    WinDrawLine(offX + widthPerSqu * 5,
                offY + heightPerSqu * 7,
                offX + widthPerSqu * 3,
                offY + heightPerSqu * 9);

    // Draw peices on offscreen window
    for(i = 0; i < BRD_SIZE; i++)
    {
        if(piece[i] == EMPTY)
        {
            continue;
        }

        bitmapH = DmGetResource(bitmapRsc, bitmapID[color[i]][piece[i]]);

        if(bitmapH)
        {
            bitmap = (BitmapPtr)MemHandleLock(bitmapH);
            MemHandleUnlock(bitmapH);

            brdX = (isFlip?(BRD_SIZE - 1 - i):i) % BRD_WIDTH;
            brdY = (isFlip?(BRD_SIZE - 1 - i):i) / BRD_WIDTH;

            // Draw now!
            WinDrawBitmap(bitmap,
                        brdX * widthPerSqu + BRD_BOUND_PIXEL,
                        brdY * heightPerSqu + BRD_BOUND_PIXEL);

            DmReleaseResource(bitmapH);
        }
    }

    // Restore and copy offscreen window to screen window
    WinSetDrawWindow(oldWin);
    RctSetRectangle(&offWinRect, 0, 0, gadgetRect.extent.x,
            gadgetRect.extent.y);
    WinCopyRectangle(offWin, NULL, &offWinRect, gadgetRect.topLeft.x,
            gadgetRect.topLeft.y, winPaint);
    WinDeleteWindow(offWin, false);

    // Restore previous drawing form
    FrmSetActiveForm(oldForm);
}

/***********************************************************************
 *
 * FUNCTION:    ScrnToBrdCoord
 *
 * DESCRIPTION: Translates screen coordinate within the boundary of gadget
 *              to XiangQi board coordinate.
 *
 * PARAMETERS:  form        FormType pointer to the form which contains
 *                          gadget.
 *              gadgetID    ID of the gadget.
 *              scrnX       Screen coordinate x.
 *              scrnY       Screen coordinate y.
 *              isFlip      Is the board already be flipped.
 *              location    Resulting board coordinate.
 *
 * RETURNED:    True if screen coordinate was translated successfully to
 *              board coordinate. Else false.
 *
 * COMMENT:     NONE
 *
 ***********************************************************************/
Boolean ScrnToBrdCoord(FormType *form, UInt16 gadgetID,
                    Int16 scrnX, Int16 scrnY, Boolean isFlip,
                    Int16 *location)
{
    UInt16 gadgetIndex;
    RectangleType gadgetRect;
    Int16 widthPerSqu;
    Int16 heightPerSqu;
    Int16 brdX;
    Int16 brdY;

    // Get the rectangle bound of gadget
    gadgetIndex = FrmGetObjectIndex(form, gadgetID);
    FrmGetObjectBounds(form, gadgetIndex, &gadgetRect);

    widthPerSqu = (gadgetRect.extent.x - BRD_BOUND_PIXEL) / (BRD_WIDTH);
    heightPerSqu = (gadgetRect.extent.y - BRD_BOUND_PIXEL) / (BRD_HEIGHT);

    // Check whether the given screen coordinate is within gadget
    if( !RctPtInRectangle(scrnX, scrnY, &gadgetRect) )
    {
        return false;
    }

    brdX = (scrnX - gadgetRect.topLeft.x) / widthPerSqu;
    brdY = (scrnY - gadgetRect.topLeft.y) / heightPerSqu;

    *location = brdY * BRD_WIDTH + brdX;

    if(isFlip)
    {
        *location =(BRD_SIZE - 1) - *location;
    }

    return (*location >= 0 && *location <= BRD_SIZE - 1);
}

/***********************************************************************
 *
 * FUNCTION:    DrawPieceBound
 *
 * DESCRIPTION: Draws a piece rectangle boundary within the selected
 *              piece.
 *
 * PARAMETERS:  form        FormType pointer to the form which contains
 *                          gadget.
 *              gadgetID    ID of the gadget.
 *              location    board coordinate of selected piece. 
 *              isFlip      Is the board already be flipped.
 *
 * RETURNED:    True if piece rectangle boundary was drawn successfully.
 *              Else false.
 *
 * COMMENT:     NONE
 *
 ***********************************************************************/
Boolean DrawPieceBound(FormType *form, UInt16 gadgetID,
                    Int16 location, Boolean isFlip)
{
    FormType *oldForm;
    UInt16 gadgetIndex;
    RectangleType gadgetRect;
    RectangleType winRect;
    Int16 widthPerSqu;
    Int16 heightPerSqu;
    Int16 brdX;
    Int16 brdY;

    // If the given board coordinate valid
    if(!(location >= 0 && location <= BRD_SIZE - 1))
    {
        return false;
    }

    // Get the current drawing form
    oldForm = FrmGetActiveForm();
    FrmSetActiveForm(form);

    // Get the rectangle bound of gadget
    gadgetIndex = FrmGetObjectIndex(form, gadgetID);
    FrmGetObjectBounds(form, gadgetIndex, &gadgetRect);

    widthPerSqu = (gadgetRect.extent.x - BRD_BOUND_PIXEL) / (BRD_WIDTH);
    heightPerSqu = (gadgetRect.extent.y - BRD_BOUND_PIXEL) / (BRD_HEIGHT);

    if(isFlip)
    {
        location = BRD_SIZE - 1 - location;
    }

    brdX = location % BRD_WIDTH;
    brdY = location / BRD_WIDTH;

    RctSetRectangle(&winRect,
                    gadgetRect.topLeft.x + brdX * widthPerSqu + BRD_BOUND_PIXEL,
                    gadgetRect.topLeft.y + brdY * heightPerSqu + BRD_BOUND_PIXEL,
                    widthPerSqu - BRD_BOUND_PIXEL,
                    heightPerSqu - BRD_BOUND_PIXEL);
    WinDrawRectangleFrame(simpleFrame, &winRect);

    // Restore previous drawing form
    FrmSetActiveForm(oldForm);

    return true;
}

/***********************************************************************
 *
 * FUNCTION:    InitBoard
 *
 * DESCRIPTION: Initialize board state.
 *
 * PARAMETERS:  NONE
 *					
 * RETURNED:    NONE
 *
 * COMMENT:     NONE
 *
 ***********************************************************************/
void InitBoard()
{
    const static Int16 color[BRD_SIZE]= {
		0, 0, 0, 0, 0, 0, 0, 0, 0,
		7, 7, 7, 7, 7, 7, 7, 7, 7,
		7, 0, 7, 7, 7, 7, 7, 0, 7,
		0, 7, 0, 7, 0, 7, 0, 7, 0,
		7, 7, 7, 7, 7, 7, 7, 7, 7,
		7, 7, 7, 7, 7, 7, 7, 7, 7,
		1, 7, 1, 7, 1, 7, 1, 7, 1,
		7, 1, 7, 7, 7, 7, 7, 1, 7,
		7, 7, 7, 7, 7, 7, 7, 7, 7,
		1, 1, 1, 1, 1, 1, 1, 1, 1};

    const static Int16 piece[BRD_SIZE]= {
		5, 3, 2, 1, 6, 1, 2, 3, 5,
		7, 7, 7, 7, 7, 7, 7, 7, 7,
		7, 4, 7, 7, 7, 7, 7, 4, 7,
		0, 7, 0, 7, 0, 7, 0, 7, 0,
		7, 7, 7, 7, 7, 7, 7, 7, 7,
		7, 7, 7, 7, 7, 7, 7, 7, 7,
		0, 7, 0, 7, 0, 7, 0, 7, 0,
		7, 4, 7, 7, 7, 7, 7, 4, 7,
		7, 7, 7, 7, 7, 7, 7, 7, 7,
		5, 3, 2, 1, 6, 1, 2, 3, 5};

    /* Counter */
    Int16 i;

    for(i = 0; i < BRD_SIZE; i++)
    {
        Color[i] = color[i];
        Piece[i] = piece[i];
    }

    Side = LIGHT;
    XSide = DARK;  
    Ply = 0;  
    HPly = 0;
    Fifty = 0;
    FirstMove[0] = 0;

    // Generate all the possible move list
    Gen();
}

/***********************************************************************
 *
 * FUNCTION:    BrdCoordToWXF
 *
 * DESCRIPTION: Converting board coordinate to WXF notation.
 *
 * PARAMETERS:  from    From board coordinate.
 *              to      To board coordinate.
 *              wxf     Holds the result of board coordinate->wxf.
 *
 * RETURNED:    NONE
 *
 * COMMENT:     The size of wxf must >= 5. Otherwise, the behaviour of
 *              this function is undetermined.
 *
 ***********************************************************************/
void BrdCoordToWXF(Int16 from, Int16 to, Char *wxf)
{
    Int16 fromBrdX;
    Int16 fromBrdY;
    Int16 toBrdX;
    Int16 toBrdY;

    fromBrdX = from % BRD_WIDTH;
    fromBrdY = from / BRD_WIDTH;
    toBrdX = to % BRD_WIDTH;
    toBrdY = to / BRD_WIDTH;

    wxf[0] = 'a' + fromBrdX;
    wxf[1] = '9' - fromBrdY;
    wxf[2] = 'a' + toBrdX;
    wxf[3] = '9' - toBrdY;
    wxf[4] = 0;
}
#endif

/***********************************************************************
 *
 * FUNCTION:    MakeMove
 *
 * DESCRIPTION: Makes a move.
 *
 * PARAMETERS:  move    Move to be made.
 *
 * RETURNED:    True if move is an illegal move. Else false.
 *
 * COMMENT:     If move is an illegal move, move will not be made.
 *
 ***********************************************************************/
Boolean MakeMove(ARMGLOBALS_PROTO MoveT move)
{
    UInt8 from;
    UInt8 to;

    from = move.b.from;
    to   = move.b.to;

    HistDat[HPly].m       = move;
    HistDat[HPly].capture = Piece[to];
    HistDat[HPly].fifty   = Fifty;

    Piece[to] = Piece[from];
    Color[to] = Side; // Color[from];
    Piece[from] = EMPTY;
    Color[from] = EMPTY;

    if( HistDat[HPly].capture == EMPTY )
    {
        Fifty++;
    }
    else
    {
        Fifty = 0;
    }

    HPly++;
    Ply++;
    Side  = Side ^ SIDE_MASK;
    XSide = Side ^ SIDE_MASK;

    // Am i in check
    if(IsInCheck(ARMGLOBALS Side ^ SIDE_MASK))
    {
        UnMakeMove(ARMGLOBALS_VOID);
        return false;
    }

    return true;
}

/***********************************************************************
 *
 * FUNCTION:    UnMakeMove
 *
 * DESCRIPTION: UnMakes a move.
 *
 * PARAMETERS:  NONE.
 *
 * RETURNED:    NONE.
 *
 * COMMENT:     Number of call on this function <= number of call on
 *              MakeMove function. Otherwise, the behaviour of this function
 *              is undetermined.
 *
 ***********************************************************************/
void UnMakeMove(ARMGLOBALS_PROTO_VOID)
{
    UInt8 from;
    UInt8 to;

    Ply--;
    HPly--;

    Side  = Side ^ SIDE_MASK;
    XSide = Side ^ SIDE_MASK;

    Fifty = HistDat[HPly].fifty;

    to   = HistDat[HPly].m.b.to;
    from = HistDat[HPly].m.b.from;

    Piece[from] = Piece[to];
    Color[from] = Side; // Color[to];
    Piece[to] = HistDat[HPly].capture;
    Color[to] = (HistDat[HPly].capture == EMPTY) ? EMPTY : XSide;
}

void MakeNullMove(ARMGLOBALS_PROTO_VOID)
{
    HistDat[HPly].m.b.to = 90;
    HistDat[HPly].m.b.from = 90;
    HistDat[HPly].capture = Piece[90];
    HistDat[HPly].fifty = Fifty;

    HPly++;
    Ply++;
    Side = Side ^ SIDE_MASK;
    XSide = Side ^ SIDE_MASK;
}

void UnMakeNullMove(ARMGLOBALS_PROTO_VOID)
{
    Ply--;
    HPly--;

    Side = Side ^ SIDE_MASK;
    XSide = Side ^ SIDE_MASK;
}
